program project1;

{$apptype console}
{$mode objfpc}{$H+}

uses
  heaptrc, sysutils, Classes;

type
  THandlerExceptionEvent = procedure (Sender: TObject; E: Exception) of object;
  THandler = class
  protected
    FRaiseException: Boolean;
    FOnException: THandlerExceptionEvent;
    procedure DoOnException(E: Exception);
    procedure InternalExecute; virtual; abstract;
  public
    procedure Assign(ASource: THandler);
    procedure Execute; virtual;
    property RaiseException: Boolean read FRaiseException write FRaiseException;
    property OnException: THandlerExceptionEvent read FOnException write FOnException;
  end;

  TLib = class(THandler)
  protected
    procedure InternalExecute; override;
  end;

  TTask = class(THandler)
  protected
    FLib: TLib;
    procedure InternalExecute; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TExceptHandler = class
  public
    procedure CatchException(Sender: TObject; E: Exception);
  end;

{ THandler }

procedure THandler.DoOnException(E: Exception);
begin
  if Assigned(FOnException) then
    FOnException(Self, E)
  else
    raise E;
end;

procedure THandler.Assign(ASource: THandler);
begin
  FRaiseException := ASource.RaiseException;
end;

procedure THandler.Execute;
begin
  try
    InternalExecute;
  except
    on E: Exception do
      DoOnException(E);
  end;
end;

{ TLib }

procedure TLib.InternalExecute;
begin
  writeln('TLib.InternalExecute');
  if FRaiseException then
    raise Exception.Create('TLib EXCEPTION!');
end;

{ TTask }

procedure TTask.InternalExecute;
begin
  FLib.Assign(Self);
  FLib.Execute;
  writeln('TTask.InternalExecute');
  if FRaiseException then
    raise Exception.Create('TTask EXCEPTION!');
end;

constructor TTask.Create;
begin
  FLib := TLib.Create;
end;

destructor TTask.Destroy;
begin
  FLib.Free;
  inherited Destroy;
end;

{ TExceptHandler }

procedure TExceptHandler.CatchException(Sender: TObject; E: Exception);
begin
  writeln(Sender.ClassName,': ', 'CatchException: ', E.Message);
end;

var
  ExceptHandler: TExceptHandler;
  Task: TTask;


begin
  ExceptHandler := TExceptHandler.Create;
  Task := TTask.Create;
  try
    try
      Task.Execute;

      Task.RaiseException := True;
      Task.OnException := @ExceptHandler.CatchException;
      Task.Execute;
    except
      on E: Exception do
        writeln('GLOBAL EXCEPTION: ' + E.Message);
    end;
  finally
    Task.Free;
    ExceptHandler.Free;
  end;

  writeln('Done');
end.

